#include <errno.h>
#include <math.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "xwiimote.h"

static struct xwii_iface *iface; //a struct containing addresses to directly communicate with differet modulues of wii remote
static int val2deg = 1.16667;

void error(const char *msg)
{
    //print error msg
    perror(msg);
    exit(0);
}

int connection(char *argv[])
{
    int sockfd, portno;
    struct sockaddr_in serv_addr;
    struct hostent *server;

    portno = atoi(argv[2]);
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0)
        error("ERROR opening socket");
    server = gethostbyname(argv[1]);
    if (server == NULL) {
        fprintf(stderr,"ERROR, no such host\n");
        exit(0);
    }
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    bcopy((char *)server->h_addr,
         (char *)&serv_addr.sin_addr.s_addr,
         server->h_length);
    serv_addr.sin_port = htons(portno);
    if (connect(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0)
    {
        error("ERROR connecting");
        return (-1);
    }
    else
        return sockfd;
}

static char *get_dev(int num)
{
    //create monitor system for wiimote device base on wiimote #
    //return path/address to the wiimote device
    struct xwii_monitor *mon;
    char *ent;
    int i = 0;

    mon = xwii_monitor_new(false, false);
    if (!mon) {
        printf("Cannot create monitor\n");
        return NULL;
    }

    while ((ent = xwii_monitor_poll(mon))) {
        if (++i == num)
            break;
        free(ent);
    }

    xwii_monitor_unref(mon);

    if (!ent)
        printf("Cannot find device with number #%d\n", num);

    return ent;
}

static void wii_event(const struct xwii_event *event, int sockfd)
{
    int n, reply[1];
    //array that stores the states of key and acceleration data
    int arg[8] = {0};

    //in event struct, v is an union that stores data payload

    //key struct continas the enumerated key number and key state
    //store key number in code and state as a bool pressed (0 false, 1 true)
    unsigned int code = event->v.key.code;
    bool pressed = event->v.key.state;

    //Use switch case to cycle through wiimote events
    switch(event->type)
    {
    case XWII_EVENT_KEY:

        //if the respective key are pressed, the state is set to 1, otherwise 0
        if(pressed)
        {
        if (code == XWII_KEY_LEFT)
            arg[2] = 1;
        else
            arg[2] = 0;
        if (code == XWII_KEY_RIGHT)
            arg[3] = 1;
        else
            arg[3] = 0;
        if (code == XWII_KEY_UP)
            arg[4] = 1;
        else
            arg[4] = 0;
        if (code == XWII_KEY_DOWN)
            arg[5] = 1;
        else
            arg[5] = 0;
        }
        else
        {
            for(int i = 2; i < 6; i++)
                arg[i] = 0;

        }

        //key A and B act as switches for performing actions, the sleep() was used to pause the program while action is performed
        if (code == XWII_KEY_A)
            if(pressed)
            {
                arg[6] = -arg[6];
                sleep(1);
            }
        if (code == XWII_KEY_B)
            if(pressed)
            {
                arg[7] = -arg[7];
                sleep(1);
            }
        break;
    case XWII_EVENT_ACCEL:

        //abs struct contains absolute motion event payload, x stores roll, while y stores pitch
        if(arg[7] == 1)
        {
            arg[0] = event->v.abs[0].x*val2deg;
            arg[1] = event->v.abs[0].y*val2deg;
        }
        else
        {
            arg[0] = 0;
            arg[1] = 0;
        }
        break;
    }

    //send motion data to server using socket
    n = write(sockfd, arg, sizeof(arg));
    if (n < 0)
         error("ERROR writing to socket");
    bzero(reply,sizeof(reply));

    n = read(sockfd,reply,sizeof(reply));
    if (n < 0)
         error("ERROR reading from socket");
}

static int run_iface(struct xwii_iface *iface, int sockfd)
{
    //wiimote notify users about any state-changes via events (data about the states of wiimote is stored in this struct)
    struct xwii_event event;
    int ret = 0, fds_num;
    //file descriptor for poll() function
    struct pollfd fds[2];

    //set file descriptor/address of poll() function to the address of wiimote
    memset(fds, 0, sizeof(fds));
    fds[0].fd = 0;
    fds[0].events = POLLIN;
    fds[1].fd = xwii_iface_get_fd(iface);
    fds[1].events = POLLIN;
    fds_num = 2;

    //watch connected wii device for hotplug events
    ret = xwii_iface_watch(iface, true);
    if (ret)
        printf("Error: Cannot initialize hotplug watch descriptor\n");

    while (1) {
        //test if there is data that can be polled from the specificed wiimote address
        ret = poll(fds, fds_num, -1);
        if (ret < 0) {
            if (errno != EINTR) {
                ret = -errno;
                printf("Error: Cannot poll fds: %d\n", ret);
                break;
            }
        }

        //read incoming event-queue from wiimote, the event is saved in event struct
        ret = xwii_iface_dispatch(iface, &event, sizeof(event));
        if (ret) {
            if (ret != -EAGAIN) {
                printf("Error: Read failed with err:%d\n",
                        ret);
                break;
            }
        }
        wii_event(&event, sockfd);

    }
    return ret;
}

int main(int argc, char **argv)
{
    int ret = 0;
    char *path = NULL;
    int sockfd;
    arg[6] = -1;
    arg[7] = -1;

    if (argc < 4) {
       fprintf(stderr,"usage %s hostname port remote#\n", argv[0]);
       exit(0);
    }
    else
    {
        //obtaining path/address to wiimote
        if (argv[3][0] != '/')
            path = get_dev(atoi(argv[3]));

        //create new device interface to communication with wiimote using address obtained
        ret = xwii_iface_new(&iface, path ? path : argv[1]);
        free(path);

        if (ret)
            printf("Cannot create xwii_iface '%s' err:%d\n", argv[1], ret);
        else
        {

            //open interfaces on this device. Interface must be opened before they can be communicated.
            //xwii-iface-available(iface) returns all available interfaces
            ret = xwii_iface_open(iface, xwii_iface_available(iface) | XWII_IFACE_WRITABLE);
            if (ret)
                printf("Error: Cannot open interface: %d",ret);

              //open acceleration interface
              ret = xwii_iface_open(iface, XWII_IFACE_ACCEL);
              if (ret)
                  printf("Error: Cannot enable accelerometer: %d\n",ret);
              else
                  printf("Info: Enable accelerometer\n");

              //open key interface
              ret = xwii_iface_open(iface, XWII_IFACE_CORE |
                               XWII_IFACE_WRITABLE);
              if (ret)
                  printf("Error: Cannot enable key events: %d\n", ret);
              else
                  printf("Info: Enable key events\n");

            //establish connection with server
            sockfd = connection(argv);
            //if connection successful, run interface communication
            if(sockfd != -1)
                ret = run_iface(iface, sockfd);
            xwii_iface_unref(iface);
            if (ret)
                printf("Program failed; press any key to exit");
        }
    }

    return abs(ret);
}
